@@ -70,7 +70,7 @@
 #include <google/protobuf/io/printer.h>
 #include <google/protobuf/io/coded_stream.h>
 #include <google/protobuf/wire_format.h>
-#include <google/protobuf/descriptor.pb.h>
+#include <protobuf-c/protobuf-c.pb.h>
 
 namespace google {
 namespace protobuf {
@@ -80,7 +80,7 @@ namespace c {
 // ===================================================================
 
 MessageGenerator::MessageGenerator(const Descriptor* descriptor,
-                                   const string& dllexport_decl)
+                                   const std::string& dllexport_decl)
   : descriptor_(descriptor),
     dllexport_decl_(dllexport_decl),
     field_generators_(descriptor),
@@ -111,8 +111,8 @@ MessageGenerator::~MessageGenerator() {}
 
 void MessageGenerator::
 GenerateStructTypedef(io::Printer* printer) {
-  printer->Print("typedef struct _$classname$ $classname$;\n",
-                 "classname", FullNameToC(descriptor_->full_name()));
+  printer->Print("typedef struct $classname$ $classname$;\n",
+                 "classname", FullNameToC(descriptor_->full_name(), descriptor_->file()));
 
   for (int i = 0; i < descriptor_->nested_type_count(); i++) {
     nested_generators_[i]->GenerateStructTypedef(printer);
@@ -137,10 +137,10 @@ GenerateStructDefinition(io::Printer* printer) {
     nested_generators_[i]->GenerateStructDefinition(printer);
   }
 
-  std::map<string, string> vars;
-  vars["classname"] = FullNameToC(descriptor_->full_name());
-  vars["lcclassname"] = FullNameToLower(descriptor_->full_name());
-  vars["ucclassname"] = FullNameToUpper(descriptor_->full_name());
+  std::map<std::string, std::string> vars;
+  vars["classname"] = FullNameToC(descriptor_->full_name(), descriptor_->file());
+  vars["lcclassname"] = FullNameToLower(descriptor_->full_name(), descriptor_->file());
+  vars["ucclassname"] = FullNameToUpper(descriptor_->full_name(), descriptor_->file());
   vars["field_count"] = SimpleItoa(descriptor_->field_count());
   if (dllexport_decl_.empty()) {
     vars["dllexport"] = "";
@@ -153,15 +153,15 @@ GenerateStructDefinition(io::Printer* printer) {
     const OneofDescriptor *oneof = descriptor_->oneof_decl(i);
     vars["opt_comma"] = ",";
 
-    vars["oneofname"] = FullNameToUpper(oneof->name());
-    vars["foneofname"] = FullNameToC(oneof->full_name());
+    vars["oneofname"] = CamelToUpper(oneof->name());
+    vars["foneofname"] = FullNameToC(oneof->full_name(), oneof->file());
 
     printer->Print("typedef enum {\n");
     printer->Indent();
     printer->Print(vars, "$ucclassname$__$oneofname$__NOT_SET = 0,\n");
     for (int j = 0; j < oneof->field_count(); j++) {
       const FieldDescriptor *field = oneof->field(j);
-      vars["fieldname"] = FullNameToUpper(field->name());
+      vars["fieldname"] = CamelToUpper(field->name());
       vars["fieldnum"] = SimpleItoa(field->number());
       bool isLast = j == oneof->field_count() - 1;
       if (isLast) {
@@ -169,7 +169,7 @@ GenerateStructDefinition(io::Printer* printer) {
       }
       printer->Print(vars, "$ucclassname$__$oneofname$_$fieldname$ = $fieldnum$$opt_comma$\n");
     }
-    printer->Print(vars, "  PROTOBUF_C__FORCE_ENUM_TO_BE_INT_SIZE($ucclassname$__$oneofname$)\n");
+    printer->Print(vars, "  PROTOBUF_C__FORCE_ENUM_TO_BE_INT_SIZE($ucclassname$__$oneofname$__CASE)\n");
     printer->Outdent();
     printer->Print(vars, "} $foneofname$Case;\n\n");
   }
@@ -178,10 +178,14 @@ GenerateStructDefinition(io::Printer* printer) {
   descriptor_->GetSourceLocation(&msgSourceLoc);
   PrintComment (printer, msgSourceLoc.leading_comments);
 
+  const ProtobufCMessageOptions opt =
+	  descriptor_->options().GetExtension(pb_c_msg);
+  vars["base"] = opt.base_field_name();
+
   printer->Print(vars,
-    "struct $dllexport$ _$classname$\n"
+    "struct $dllexport$ $classname$\n"
     "{\n"
-    "  ProtobufCMessage base;\n");
+    "  ProtobufCMessage $base$;\n");
 
   // Generate fields.
   printer->Indent();
@@ -200,8 +204,8 @@ GenerateStructDefinition(io::Printer* printer) {
   // Generate unions from oneofs.
   for (int i = 0; i < descriptor_->oneof_decl_count(); i++) {
     const OneofDescriptor *oneof = descriptor_->oneof_decl(i);
-    vars["oneofname"] = FullNameToLower(oneof->name());
-    vars["foneofname"] = FullNameToC(oneof->full_name());
+    vars["oneofname"] = CamelToLower(oneof->name());
+    vars["foneofname"] = FullNameToC(oneof->full_name(), oneof->file());
 
     printer->Print(vars, "$foneofname$Case $oneofname$_case;\n");
 
@@ -241,7 +245,7 @@ GenerateStructDefinition(io::Printer* printer) {
   }
   for (int i = 0; i < descriptor_->oneof_decl_count(); i++) {
     const OneofDescriptor *oneof = descriptor_->oneof_decl(i);
-    vars["foneofname"] = FullNameToUpper(oneof->full_name());
+    vars["foneofname"] = FullNameToUpper(oneof->full_name(), oneof->file());
     // Initialize the case enum
     printer->Print(vars, ", $foneofname$__NOT_SET");
     // Initialize the union
@@ -252,21 +256,38 @@ GenerateStructDefinition(io::Printer* printer) {
 }
 
 void MessageGenerator::
-GenerateHelperFunctionDeclarations(io::Printer* printer, bool is_submessage)
+GenerateHelperFunctionDeclarations(io::Printer* printer,
+				   bool is_pack_deep,
+				   bool gen_pack,
+				   bool gen_init)
 {
+  const ProtobufCMessageOptions opt =
+	  descriptor_->options().GetExtension(pb_c_msg);
+
+  // Override parent settings, if needed
+  if (opt.has_gen_pack_helpers())
+    gen_pack = opt.gen_pack_helpers();
+  if (opt.has_gen_init_helpers())
+    gen_init = opt.gen_init_helpers();
+
   for (int i = 0; i < descriptor_->nested_type_count(); i++) {
-    nested_generators_[i]->GenerateHelperFunctionDeclarations(printer, true);
+    bool nested_pack = !is_pack_deep ? opt.gen_pack_helpers() : gen_pack;
+    nested_generators_[i]->GenerateHelperFunctionDeclarations(printer, true,
+							      nested_pack,
+							      gen_init);
   }
 
-  std::map<string, string> vars;
-  vars["classname"] = FullNameToC(descriptor_->full_name());
-  vars["lcclassname"] = FullNameToLower(descriptor_->full_name());
-  printer->Print(vars,
+  std::map<std::string, std::string> vars;
+  vars["classname"] = FullNameToC(descriptor_->full_name(), descriptor_->file());
+  vars["lcclassname"] = FullNameToLower(descriptor_->full_name(), descriptor_->file());
+  if (gen_init) {
+    printer->Print(vars,
 		 "/* $classname$ methods */\n"
 		 "void   $lcclassname$__init\n"
 		 "                     ($classname$         *message);\n"
 		);
-  if (!is_submessage) {
+  }
+  if (gen_pack) {
     printer->Print(vars,
 		 "size_t $lcclassname$__get_packed_size\n"
 		 "                     (const $classname$   *message);\n"
@@ -291,7 +312,7 @@ GenerateHelperFunctionDeclarations(io::Printer* printer, bool is_submessage)
 void MessageGenerator::
 GenerateDescriptorDeclarations(io::Printer* printer) {
   printer->Print("extern const ProtobufCMessageDescriptor $name$__descriptor;\n",
-                 "name", FullNameToLower(descriptor_->full_name()));
+                 "name", FullNameToLower(descriptor_->full_name(), descriptor_->file()));
 
   for (int i = 0; i < descriptor_->nested_type_count(); i++) {
     nested_generators_[i]->GenerateDescriptorDeclarations(printer);
@@ -306,8 +327,8 @@ void MessageGenerator::GenerateClosureTypedef(io::Printer* printer)
   for (int i = 0; i < descriptor_->nested_type_count(); i++) {
     nested_generators_[i]->GenerateClosureTypedef(printer);
   }
-  std::map<string, string> vars;
-  vars["name"] = FullNameToC(descriptor_->full_name());
+  std::map<std::string, std::string> vars;
+  vars["name"] = FullNameToC(descriptor_->full_name(), descriptor_->file());
   printer->Print(vars,
                  "typedef void (*$name$_Closure)\n"
 		 "                 (const $name$ *message,\n"
@@ -325,43 +346,61 @@ compare_pfields_by_number (const void *a, const void *b)
 }
 
 void MessageGenerator::
-GenerateHelperFunctionDefinitions(io::Printer* printer, bool is_submessage)
+GenerateHelperFunctionDefinitions(io::Printer* printer,
+				  bool is_pack_deep,
+				  bool gen_pack,
+				  bool gen_init)
 {
+  const ProtobufCMessageOptions opt =
+	  descriptor_->options().GetExtension(pb_c_msg);
+
+  // Override parent settings, if needed
+  if (opt.has_gen_pack_helpers())
+    gen_pack = opt.gen_pack_helpers();
+  if (opt.has_gen_init_helpers())
+    gen_init = opt.gen_init_helpers();
+
   for (int i = 0; i < descriptor_->nested_type_count(); i++) {
-    nested_generators_[i]->GenerateHelperFunctionDefinitions(printer, true);
+    bool nested_pack = !is_pack_deep ? opt.gen_pack_helpers() : gen_pack;
+    nested_generators_[i]->GenerateHelperFunctionDefinitions(printer, true,
+							     nested_pack,
+							     gen_init);
   }
 
-  std::map<string, string> vars;
-  vars["classname"] = FullNameToC(descriptor_->full_name());
-  vars["lcclassname"] = FullNameToLower(descriptor_->full_name());
-  vars["ucclassname"] = FullNameToUpper(descriptor_->full_name());
-  printer->Print(vars,
+  std::map<std::string, std::string> vars;
+  vars["classname"] = FullNameToC(descriptor_->full_name(), descriptor_->file());
+  vars["lcclassname"] = FullNameToLower(descriptor_->full_name(), descriptor_->file());
+  vars["ucclassname"] = FullNameToUpper(descriptor_->full_name(), descriptor_->file());
+  vars["base"] = opt.base_field_name();
+  if (gen_init) {
+    printer->Print(vars,
 		 "void   $lcclassname$__init\n"
 		 "                     ($classname$         *message)\n"
 		 "{\n"
 		 "  static const $classname$ init_value = $ucclassname$__INIT;\n"
 		 "  *message = init_value;\n"
 		 "}\n");
-  if (!is_submessage) {
+  }
+  if (gen_pack) {
     printer->Print(vars,
 		 "size_t $lcclassname$__get_packed_size\n"
 		 "                     (const $classname$ *message)\n"
 		 "{\n"
-		 "  assert(message->base.descriptor == &$lcclassname$__descriptor);\n"
+		 "  assert(message->$base$.descriptor == &$lcclassname$__descriptor);\n"
 		 "  return protobuf_c_message_get_packed_size ((const ProtobufCMessage*)(message));\n"
 		 "}\n"
 		 "size_t $lcclassname$__pack\n"
 		 "                     (const $classname$ *message,\n"
 		 "                      uint8_t       *out)\n"
 		 "{\n"
-		 "  assert(message->base.descriptor == &$lcclassname$__descriptor);\n"
+		 "  assert(message->$base$.descriptor == &$lcclassname$__descriptor);\n"
 		 "  return protobuf_c_message_pack ((const ProtobufCMessage*)message, out);\n"
 		 "}\n"
 		 "size_t $lcclassname$__pack_to_buffer\n"
 		 "                     (const $classname$ *message,\n"
 		 "                      ProtobufCBuffer *buffer)\n"
 		 "{\n"
-		 "  assert(message->base.descriptor == &$lcclassname$__descriptor);\n"
+		 "  assert(message->$base$.descriptor == &$lcclassname$__descriptor);\n"
 		 "  return protobuf_c_message_pack_to_buffer ((const ProtobufCMessage*)message, buffer);\n"
 		 "}\n"
 		 "$classname$ *\n"
@@ -380,7 +419,7 @@ GenerateHelperFunctionDefinitions(io::Printer* printer, bool is_submessage)
 		 "{\n"
 		 "  if(!message)\n"
 		 "    return;\n"
-		 "  assert(message->base.descriptor == &$lcclassname$__descriptor);\n"
+		 "  assert(message->$base$.descriptor == &$lcclassname$__descriptor);\n"
 		 "  protobuf_c_message_free_unpacked ((ProtobufCMessage*)message, allocator);\n"
 		 "}\n"
 		);
@@ -388,11 +427,11 @@ GenerateHelperFunctionDefinitions(io::Printer* printer, bool is_submessage)
 }
 
 void MessageGenerator::
-GenerateMessageDescriptor(io::Printer* printer) {
-    std::map<string, string> vars;
+GenerateMessageDescriptor(io::Printer* printer, bool gen_init) {
+    std::map<std::string, std::string> vars;
     vars["fullname"] = descriptor_->full_name();
-    vars["classname"] = FullNameToC(descriptor_->full_name());
-    vars["lcclassname"] = FullNameToLower(descriptor_->full_name());
+    vars["classname"] = FullNameToC(descriptor_->full_name(), descriptor_->file());
+    vars["lcclassname"] = FullNameToLower(descriptor_->full_name(), descriptor_->file());
     vars["shortname"] = ToCamel(descriptor_->name());
     vars["n_fields"] = SimpleItoa(descriptor_->field_count());
     vars["packagename"] = descriptor_->file()->package();
@@ -401,8 +440,14 @@ GenerateMessageDescriptor(io::Printer* printer) {
         descriptor_->file()->options().optimize_for() ==
         FileOptions_OptimizeMode_CODE_SIZE;
 
+    const ProtobufCMessageOptions opt =
+	    descriptor_->options().GetExtension(pb_c_msg);
+    // Override parent settings, if needed
+    if (opt.has_gen_init_helpers())
+      gen_init = opt.gen_init_helpers();
+
     for (int i = 0; i < descriptor_->nested_type_count(); i++) {
-      nested_generators_[i]->GenerateMessageDescriptor(printer);
+      nested_generators_[i]->GenerateMessageDescriptor(printer, gen_init);
     }
 
     for (int i = 0; i < descriptor_->enum_type_count(); i++) {
@@ -418,6 +463,7 @@ GenerateMessageDescriptor(io::Printer* printer) {
 
     for (int i = 0; i < descriptor_->field_count(); i++) {
       const FieldDescriptor *fd = descriptor_->field(i);
+      const ProtobufCFieldOptions opt = fd->options().GetExtension(pb_c_field);
       if (fd->has_default_value()) {
 
 	bool already_defined = false;
@@ -453,10 +499,10 @@ GenerateMessageDescriptor(io::Printer* printer) {
 	  // NOTE: not supported by protobuf
 	  vars["maybe_static"] = "";
 	  vars["field_dv_ctype"] = "{ ... }";
-	  GOOGLE_LOG(DFATAL) << "Messages can't have default values!";
+	  GOOGLE_LOG(FATAL) << "Messages can't have default values!";
 	  break;
 	case FieldDescriptor::CPPTYPE_STRING:
-	  if (fd->type() == FieldDescriptor::TYPE_BYTES)
+	  if (fd->type() == FieldDescriptor::TYPE_BYTES || opt.string_as_bytes())
 	  {
 	    vars["field_dv_ctype"] = "ProtobufCBinaryData";
 	  }
@@ -471,11 +517,11 @@ GenerateMessageDescriptor(io::Printer* printer) {
 	case FieldDescriptor::CPPTYPE_ENUM:
 	  {
 	    const EnumValueDescriptor *vd = fd->default_value_enum();
-	    vars["field_dv_ctype"] = FullNameToC(vd->type()->full_name());
+	    vars["field_dv_ctype"] = FullNameToC(vd->type()->full_name(), vd->type()->file());
 	    break;
 	  }
 	default:
-	  GOOGLE_LOG(DFATAL) << "Unknown CPPTYPE";
+	  GOOGLE_LOG(FATAL) << "Unknown CPPTYPE";
 	  break;
 	}
 	if (!already_defined)
@@ -568,8 +614,15 @@ GenerateMessageDescriptor(io::Printer* printer) {
   }
   printer->Print(vars,
       "  $n_ranges$,"
-      "  $lcclassname$__number_ranges,\n"
-      "  (ProtobufCMessageInit) $lcclassname$__init,\n"
+      "  $lcclassname$__number_ranges,\n");
+  if (gen_init) {
+    printer->Print(vars,
+      "  (ProtobufCMessageInit) $lcclassname$__init,\n");
+  } else {
+    printer->Print(vars,
+      "  NULL, /* gen_init_helpers = false */\n");
+  }
+  printer->Print(vars,
       "  NULL,NULL,NULL    /* reserved[123] */\n"
       "};\n");
 }
